from .tree.tree_parser import TreeParser

import inspect
import os

def initializer_error_handler(func):
    def wrapped_init(*args, **kwargs):
        
        try:
            func(*args, **kwargs)
            
        except Exception as e:
            
            sub_str = 'takes 1 positional argument but'
            
            if sub_str in str(e):
                raise Exception('Argteller disallows positional arguments. Please only use keyword arguments.')
                
            else:
                raise type(e)(str(e))
            
    return wrapped_init


class ArgtellerClassDecorator():
    
    def __init__(self, map_str, override=False):
        
        if os.path.exists(map_str):
            with open(map_str) as f:
                self.map_str = f.read()
                
        self.override = override
        
    def __call__(self, cls):
        
        class Wrapped(cls):
            
            @initializer_error_handler
            def __init__(cls_self, **kwargs):


                self.tree_parser = TreeParser()
                self.arg_tree = self.tree_parser.parse_tree(self.map_str)  # preset values set here


                tmp_cls_obj = cls.__new__(cls)
                names, _, _, defaults, _, _, _ = inspect.getfullargspec(tmp_cls_obj.__init__)

                if defaults is not None:
                    for name, default in zip(reversed(names), reversed(defaults)):

                        if not hasattr(cls_self, name):
                            setattr(cls_self, name, default)
                
                # set node values
                for k, v in kwargs.items():
                    
                    if k in self.tree_parser.param_nodes_dict:
                        for node in self.tree_parser.param_nodes_dict[k]:
                            node.set_value(v)

                # for optional arguments
                for k, v in kwargs.items():
                    
                    if k in self.tree_parser.option_nodes_dict:

                        for node in self.tree_parser.option_nodes_dict[k]:
                            node.set_value(v)


                self.arg_tree.traverse_tree_find_used_params()


                for name, preset_value in self.arg_tree.messenger_node_pre.used_conditional_params:

                    if preset_value is None or name in kwargs:
                        continue

                    for node in self.tree_parser.param_nodes_dict[name]:

                        # print(name, preset_value)

                        node.value = preset_value

                    for node in self.tree_parser.option_nodes_dict[name]:

                        node.value = preset_value

                    preset_value = None


                # set cls fields
                for k, v in self.tree_parser.param_nodes_dict.items():
                    
                    if self.override:
                        setattr(cls_self, k, v[0].value)
                        
                    else:
                        
                        if not hasattr(cls_self, k):
                            setattr(cls_self, k, v[0].value)

                for k, v in self.tree_parser.option_nodes_dict.items():
                    
                    if self.override:
                        setattr(cls_self, k, v[0].value)
                        
                    else:
                        
                        if not hasattr(cls_self, k):
                            setattr(cls_self, k, v[0].value)


                self.arg_tree.traverse_tree_tell()

                if self.arg_tree.total_num_missing_args > 0:

                    self.tree_parser.tree.reset_param_node_values()
                    self.arg_tree.total_num_missing_args = 0  # must reset 
                    
                    return

                tmp_kwargs = {k: v for k, v in kwargs.items() if k in names[1:]}
                super(Wrapped, cls_self).__init__(**tmp_kwargs)

                self.tree_parser.tree.reset_param_node_values()
                
            cls.injected_method = self.injected_method
            
        return Wrapped
            
    def injected_method(self, a):
        return(a)
        
# @ArgtellerClassDecorator(map_str=string)
# class A():
    
#     def __init__(self, a=None):
#         self.a = a
#         pass
    



from .nodes import ArgTree
from .nodes import TopicNode
from .nodes import ParamNode
from .nodes import AvailNode
from .nodes import ExampleNode
from .nodes import CondNode
# from .nodes import OptionNode

from collections import defaultdict
import re

class TreeParser():


    def __init__(self):

        self.topic_nodes_dict = dict()
        self.param_nodes_dict = defaultdict(list)
        self.option_nodes_dict = defaultdict(list)
        self.preset_params = list()

        self.tree = None

    def parse_tree(self, string):
        
        current_topic = None
        current_param = None

        self.tree = ArgTree(name='tree', depth=-2)

        node_count = 0
        current_pos_dict = {}



        for line in string.splitlines():

            tab_depth = line.count('\t')
            space_depth = int((len(line) - len(line.lstrip(' ')))/4)
            depth = tab_depth + space_depth

            line = line.strip()

            if line=='':
                continue
            if line[0] == '-':
                node_type = 'param'
            elif line[0] == '+':
                node_type = 'option'
            elif line[0:2] == '==':
                node_type = 'example'
            elif line[0] == '=':
                node_type = 'avail'
            elif line[0] == '?':
                node_type = 'cond'
            else:
                node_type = 'topic'


            name = re.sub('^[\s=+-?]+', '', line)


            # print(name, depth, tab_depth, space_depth)


            if node_type=='topic':

                if name != current_topic:
                    current_topic = name
                    topic_node = TopicNode(name, depth=-1)
                    self.tree.add_topic(topic_node)

                    current_pos_dict[-1] = topic_node

                    self.topic_nodes_dict[name] = topic_node

            elif node_type=='param':


                if ':' in name:
                    

                    name, preset_value = name.split(':')
                    preset_value = eval(preset_value)
                    self.preset_params.append(name)
                

                else:
                    preset_value = None


                current_param = name

                prev_node = current_pos_dict[depth-1]

                current_node = ParamNode(name, depth)
                prev_node.add_param(current_node)


                if preset_value is not None:

                    current_node.set_preset_value(preset_value)
                    current_node.make_preset()
                    preset_value = None

                current_pos_dict[depth] = current_node


                self.param_nodes_dict[name].append(current_node)

            elif node_type=='cond':


                if name in self.param_nodes_dict:

                    for node in self.param_nodes_dict[name]:

                        preset_value = node.preset_value
                        

                current_param = name

                prev_node = current_pos_dict[depth-1]

                current_node = CondNode(name, depth)
                prev_node.add_param(current_node)

                current_pos_dict[depth] = current_node

                if preset_value is not None:

                    current_node.set_preset_value(preset_value)
                    preset_value = None


                self.param_nodes_dict[name].append(current_node)


            elif node_type=='avail':

                prev_node = current_pos_dict[depth-1]

                current_node = AvailNode(name, depth)
                current_node.set_param(current_param)

                prev_node.add_avail(current_node)

                current_pos_dict[depth] = current_node

            elif node_type=='example':

                prev_node = current_pos_dict[depth-1]

                current_node = ExampleNode(name, depth)
                prev_node.add_example(current_node)

                current_pos_dict[depth] = current_node

            elif node_type=='option':  # only valid until here, afterwards, node type is param

                prev_node = current_pos_dict[depth-1]

                current_node = ParamNode(name, depth)
                prev_node.add_option(current_node)

                current_pos_dict[depth] = current_node

                self.option_nodes_dict[name].append(current_node)
                
        return self.tree















from .base_node import BaseNode

import collections

class MessengerNode():

    def __init__(self, fail_not_messaged=None, used_conditional_params=None):

        self.fail_not_messaged = fail_not_messaged
        self.used_conditional_params = used_conditional_params

class ArgTree(BaseNode):
    
    def __init__(self, name=None, depth=None):
        super(ArgTree, self).__init__(name, depth, node_type='tree')

        self.total_num_missing_args = 0
        self.tested_topics = []
        self.used_conditional_params = []

    def traverse_tree_find_used_params(self):

        self.used_conditional_params = []

        self.messenger_node_pre = MessengerNode(used_conditional_params=self.used_conditional_params)

        ArgTree._traverse_tree_find_used_params(self, self.messenger_node_pre)

    @staticmethod
    def _traverse_tree_find_used_params(self, messenger_node):

        if self.has_topics():  # perhaps change it to checking the type of the node for stronger contract

            for topic in self.topics:
                
                ArgTree._traverse_tree_find_used_params(topic, messenger_node=messenger_node)
                    
        if self.has_params() and (self.node_type=='topic' or self.node_type=='avail' or 
            (self.node_type=='cond' and self.value==True)):

            missing_required_arguments = []
            for param in self.params:
                if param.value is None:
                    missing_required_arguments.append(param.name)

            if len(missing_required_arguments)>0 and self.node_type=='avail':

                for param in self.params:

                    messenger_node.used_conditional_params.append([param.name, param.preset_value])

            for param in self.params:

                if param.name in missing_required_arguments:
                
                    ArgTree._traverse_tree_find_used_params(param, messenger_node=messenger_node)

                else:
                    ArgTree._traverse_tree_find_used_params(param, messenger_node=messenger_node)
                
        if self.has_avails() and self.node_type=='param':
            
            for avail in self.avails:
                
                if self.value == avail.name:  # checks the avails!

                    ArgTree._traverse_tree_find_used_params(avail, messenger_node=messenger_node)

        if self.has_examples() and self.node_type=='param':
            
            for example in self.examples:

                ArgTree._traverse_tree_find_used_params(example, messenger_node=messenger_node)

        if self.has_options():
            
            for option in self.options:
                
                ArgTree._traverse_tree_find_used_params(option, messenger_node=messenger_node)

    def traverse_tree_tell(self):
        
        self.messenger_node = MessengerNode(used_conditional_params=self.used_conditional_params)

        ArgTree._traverse_tree_tell(self, False, self.messenger_node)
        ArgTree._traverse_tree_tell_options(self, self.messenger_node)

    @staticmethod
    def _traverse_tree_tell_options(self, messenger_node=None):

        if self.has_topics():

            for topic in self.topics:

                if topic not in self.tested_topics:
                    return

                ArgTree._traverse_tree_tell_options(topic, messenger_node)

        if self.has_params() and (self.node_type=='topic' or self.node_type=='avail'):

            missing_required_arguments = []
            for param in self.params:
                if param.value is None:
                    missing_required_arguments.append(param.name)

            for param in self.params:

                ArgTree._traverse_tree_tell_options(param, messenger_node)

        if self.has_avails() and self.node_type=='param':

            for avail in self.avails:
                
                ArgTree._traverse_tree_tell_options(avail, messenger_node)

        if self.has_examples() and self.node_type=='param':
            
            for example in self.examples:

                ArgTree._traverse_tree_tell_options(example, messenger_node)

        if self.has_options():

            missing_optional_arguments = []

            for option in self.options:

                used_param_names = [elem[0] for elem in messenger_node.used_conditional_params]

                if option.name in used_param_names:
                    continue

                if option.value is None:
                    missing_optional_arguments.append(option.name)

            if len(missing_optional_arguments) > 0:

                print('\nOptional argument(s) for {}:\n\n\u25BA {}'.format(self.name, '  '.join(missing_optional_arguments)))


            for option in self.options:

                ArgTree._traverse_tree_tell_options(option, messenger_node)

    @staticmethod
    def _traverse_tree_tell(self, missing=False, messenger_node=None):

        num_missing_args = 0
        
        
        if self.has_topics():  # perhaps change it to checking the type of the node for stronger contract
        
            for topic in self.topics:

                self.tested_topics.append(topic)

                print("\u2714 Checking {} requirements...     ".format(
                    topic.name), end="", flush=True)
                
                messenger_node.fail_not_messaged=True
                num_missing_args = ArgTree._traverse_tree_tell(topic, messenger_node=messenger_node)

                self.total_num_missing_args += num_missing_args

                if num_missing_args > 0:

                    return 

                else:
                    print('Passed!')
                    
                
        if self.has_params() and (self.node_type=='topic' or self.node_type=='avail' or 
            (self.node_type=='cond' and self.value==True)):

            num_missing_args = 0

            missing_required_arguments = []
            for param in self.params:
                if param.value is None:
                    missing_required_arguments.append(param.name)


            num_missing_args += len(missing_required_arguments)

            if len(missing_required_arguments)>0 and self.node_type=='topic':

                if messenger_node.fail_not_messaged:
                    print("Failed!")
                    messenger_node.fail_not_messaged = False

                print('\nRequired argument(s):\n\n\u25BA {}'.format('  '.join(missing_required_arguments)))

            elif len(missing_required_arguments)>0 and self.node_type=='cond':

                if messenger_node.fail_not_messaged:
                    print("Failed!")
                    messenger_node.fail_not_messaged = False

                print('\nRequired argument(s) for [ {} ] option:\n\n\u25BA {}'.format(self.name, '  '.join(missing_required_arguments)))

            elif len(missing_required_arguments)>0 and self.node_type=='avail':

                if messenger_node.fail_not_messaged:
                    print("Failed!")
                    messenger_node.fail_not_messaged = False

                print('\nRequired argument(s) for [ {} ] {}:\n\n\u25BA {}'.format(
                    self.name, self.param, '  '.join(missing_required_arguments)))

            for param in self.params:

                num_missing_args_from_below = 0

                if param.name in missing_required_arguments:
                
                    num_missing_args_from_below = ArgTree._traverse_tree_tell(param, True, messenger_node=messenger_node)

                else:
                    num_missing_args_from_below = ArgTree._traverse_tree_tell(param, False, messenger_node=messenger_node)


                if num_missing_args_from_below is None:

                    num_missing_args_from_below = 0

                num_missing_args += num_missing_args_from_below

        if self.has_avails() and self.node_type=='param':

            available_arguments = []

            num_missing_args = 0
            
            for avail in self.avails:

                available_arguments.append(avail.name)
                
                if self.value == avail.name:  # checks the avails!

                    num_missing_args_from_below = ArgTree._traverse_tree_tell(avail, messenger_node=messenger_node)

                    if num_missing_args_from_below is None:
                        num_missing_args_from_below = 0

                    num_missing_args += num_missing_args_from_below

            if missing:
                print('\nAvailable [ {} ] options:\n\n'
                      '\u25BA {}'.format(self.name, '  '.join(available_arguments)))

            return num_missing_args

        if self.has_examples() and self.node_type=='param':
            
            for example in self.examples:

                if missing:
                    print('\nExamples for [ {} ]: {}'.format(self.name, example.name))

                ArgTree._traverse_tree_tell(example, messenger_node=messenger_node)
                
        if self.has_options():
            
            for option in self.options:
                
                ArgTree._traverse_tree_tell(option, messenger_node=messenger_node)

        if num_missing_args is not None:
            return num_missing_args
        else:
            return 0
        
class TopicNode(BaseNode):
    
    def __init__(self, name, depth):
        super(TopicNode, self).__init__(name, depth, node_type='topic')
        
class ParamNode(BaseNode):
    
    def __init__(self, name, depth):
        super(ParamNode, self).__init__(name, depth, node_type='param')

        self._is_preset = False
        self.preset_value = None

    def set_value(self, value):

        self.value = value

    def has_value(self):

        return self.value is None

    def set_preset_value(self, preset_value):

        self.preset_value = preset_value

    def has_preset_value(self):

        return self.preset_value is None

    def make_preset(self):

        self._is_preset = True

    def is_preset(self):

        return self._is_preset

class CondNode(BaseNode):
    
    def __init__(self, name, depth):
        super(CondNode, self).__init__(name, depth, node_type='cond')

    def set_value(self, value):

        self.value = value

    def has_value(self):

        return self.value is None
        

class AvailNode(BaseNode):
    
    def __init__(self, name, depth):
        super(AvailNode, self).__init__(name, depth, node_type='avail')

        self.param = None

    def set_param(self, value):

        self.param = value

    def has_param(self):

        return self.param is not None
        
class ExampleNode(BaseNode):
    
    def __init__(self, name, depth):
        super(ExampleNode, self).__init__(name, depth, node_type='example')






class BaseNode():
    
    def __init__(self, name, depth, node_type):
        
        self.name = name
        self.node_type = node_type  # topic, param, 
        self.depth = depth

        self.value = None
        
        self.topics = []
        self.params = []
        self.avails = []
        self.examples = []
        self.options = []
        
    def add_topic(self, topic_node):
        self.topics.append(topic_node)
        
    def has_topics(self):
        return len(self.topics)>0
        
    def add_param(self, param_node):
        self.params.append(param_node)
        
    def has_params(self):
        return len(self.params)>0
        
    def add_avail(self, avail_node):
        self.avails.append(avail_node)
        
    def has_avails(self):
        return len(self.avails)>0
    
    def add_example(self, example_node):
        self.examples.append(example_node)
        
    def has_examples(self):
        return len(self.examples)>0
    
    def add_option(self, option_node):
        self.options.append(option_node)
        
    def has_options(self):
        return len(self.options)>0

    def reset_param_node_values(self):

        def applied_method(node):
            if node.node_type=='param' and not node.is_preset():
                node.value=None

        self.traverse_tree_apply(applied_method) 

    
    def traverse_tree_apply(self, method):

        BaseNode._traverse_tree_apply(self, method)


    @staticmethod
    def _traverse_tree_apply(self, method):

        if self.has_topics():
        
            for topic in self.topics:
            
                method(topic)
                
                BaseNode._traverse_tree_apply(topic, method)
                
        if self.has_params():
            
            for param in self.params:
                
                method(param)
                
                BaseNode._traverse_tree_apply(param, method)
                
        if self.has_avails():
            
            for avail in self.avails:
                
                method(avail)
                
                BaseNode._traverse_tree_apply(avail, method)
                
        if self.has_examples():
            
            for example in self.examples:
                
                method(example)
                
                BaseNode._traverse_tree_apply(example, method)
                
        if self.has_options():
            
            for option in self.options:
                
                method(option)
                
                BaseNode._traverse_tree_apply(option, method)



    def traverse_tree_print(self):
        
        BaseNode._traverse_tree_print(self)
    
    @staticmethod
    def _traverse_tree_print(self):
        
        if self.has_topics():
        
            for topic in self.topics:
            
                print('\n{}'.format(topic.name))
                
                BaseNode._traverse_tree_print(topic)
                
        if self.has_params():
            
            for param in self.params:
                
                print('{}-{} : {}'.format(' '*param.depth*4, param.name, param.value))
                
                BaseNode._traverse_tree_print(param)
                
        if self.has_avails():
            
            for avail in self.avails:
                
                print('{}={}'.format(' '*avail.depth*4, avail.name))
                
                BaseNode._traverse_tree_print(avail)
                
        if self.has_examples():
            
            for example in self.examples:
                
                print('{}=={}'.format(' '*example.depth*4, example.name))
                
                BaseNode._traverse_tree_print(example)
                
        if self.has_options():
            
            for option in self.options:
                
                print('{}+{}'.format(' '*option.depth*4, option.name))
                
                BaseNode._traverse_tree_print(option)



                